@@ -1,8 +1,8 @@
 package Apache::Options;
-require Apache::Constants;
-use vars qw(@ISA @EXPORT);
+use Apache::Constants ();
 @ISA = qw(Apache::Constants);
-@EXPORT = @{$Apache::Constants::EXPORT_TAGS{options}};
+*EXPORT = $Apache::Constants::EXPORT_TAGS{options};
+$VERSION = (qw$Revision: 1.16 $)[1];
 
 1;
 
