@@ -1,3 +1,10 @@
+#ifdef WIN32
+#define NO_PERL_SECTIONS
+#define NO_PERL_CHILD_INIT
+#define NO_PERL_CHILD_EXIT
+#include "dirent.h"
+#endif
+
 #include "EXTERN.h"
 #include "perl.h"
 #include "XSUB.h"
@@ -6,6 +13,9 @@
  * expanded to Perl_foo
  * but some cause conflict when expanded in other headers files
  */
+#undef S_ISREG
+#undef DIR
+#undef VOIDUSED
 #undef pregexec
 #undef pregfree
 #undef pregcomp
@@ -14,6 +24,7 @@
 #undef sync
 #undef my_memcmp
 #undef RETURN
+#undef die
 
 #include "httpd.h"
 #include "http_config.h"
@@ -26,15 +37,20 @@
 #include "http_conf_globals.h"
 
 typedef request_rec * Apache;
+typedef request_rec * Apache__SubRequest;
 typedef conn_rec    * Apache__Connection;
 typedef server_rec  * Apache__Server;
 
+#define GvHV_init(gv) (void)gv_fetchpv(gv, GV_ADDMULTI, SVt_PVHV)
+
 #define iniHV(hv) hv = (HV*)sv_2mortal((SV*)newHV())
 #define iniAV(av) av = (AV*)sv_2mortal((SV*)newAV())
 
+#define AvTRUE(av) (av && (AvFILL(av) > -1) && SvREFCNT(av))
+
 #define av_copy_array(av) av_make(av_len(av)+1, AvARRAY(av))  
 
-#define PerlEnvHV GvHV(gv_fetchpv("ENV", FALSE, SVt_PVHV))
+#define PerlEnvHV GvHV(envgv)
 
 #ifndef newRV_noinc
 #define newRV_noinc(sv)	((Sv = newRV(sv)), --SvREFCNT(SvRV(Sv)), Sv)
@@ -47,19 +63,68 @@ typedef server_rec  * Apache__Server;
 #define HV_SvTAINTED_on(hv,key,klen) \
     SvTAINTED_on(*hv_fetch(hv, key, klen, 0)) 
 
-/* flush %ENV */
-#define perl_clear_env \
-      hv_clear(PerlEnvHV)
-
-#define perl_set_pid \
-      if(!set_pid++) sv_setiv(GvSV(gv_fetchpv("$", TRUE, SVt_PV)), (I32)getpid())
-
 #ifdef PERL_TRACE
 #define MP_TRACE(a) a 
 #else
 #define MP_TRACE(a)
 #endif
 
+/* cut down on some noise in source */
+#define dSTATUS int status = DECLINED
+
+#define dPPDIR \
+   perl_dir_config *cld = get_module_config(r->per_dir_config, &perl_module)   
+
+#define dPSRV(srv) \
+   perl_server_config *cls = get_module_config (srv->module_config, &perl_module)
+
+/* per-directory flags */
+
+#define MPf_INCPUSH	0x00000100 /* use lib split ":", $ENV{PERL5LIB} */
+#define MPf_SENDHDR	0x00000200 /* is PerlSendHeader On? */
+#define MPf_SENTHDR	0x00000400 /* has PerlSendHeader sent the headers? */
+#define MPf_ENV		0x00000800 /* PerlSetupEnv */
+#define MPf_HASENV	0x00001000 /* do we have any PerlSetEnv's? */
+#define MPf_DSTDERR	0x00002000 /* redirect stderr to error_log */
+#define MPf_CLEANUP	0x00004000 /* did we register our cleanup ? */
+#define MPf_RCLEANUP	0x00008000 /* for $r->register_cleanup */
+
+#define MP_FMERGE(new,add,base,f) \
+if((add->flags & f) || (base->flags & f)) \
+    new->flags |= f
+    
+#define MP_INCPUSH(d)    (d->flags & MPf_INCPUSH)
+#define MP_INCPUSH_on(d)  (d->flags |= MPf_INCPUSH)
+#define MP_INCPUSH_off(d)  (d->flags  &= ~MPf_INCPUSH)
+
+#define MP_SENDHDR(d)    (d->flags & MPf_SENDHDR)
+#define MP_SENDHDR_on(d)  (d->flags |= MPf_SENDHDR)
+#define MP_SENDHDR_off(d)  (d->flags  &= ~MPf_SENDHDR)
+
+#define MP_SENTHDR(d)    (d->flags & MPf_SENTHDR)
+#define MP_SENTHDR_on(d)  (d->flags |= MPf_SENTHDR)
+#define MP_SENTHDR_off(d)  (d->flags  &= ~MPf_SENTHDR)
+
+#define MP_ENV(d)       (d->flags & MPf_ENV)
+#define MP_ENV_on(d)     (d->flags |= MPf_ENV)
+#define MP_ENV_off(d)    (d->flags  &= ~MPf_ENV)
+
+#define MP_HASENV(d)    (d->flags & MPf_HASENV)
+#define MP_HASENV_on(d)  (d->flags |= MPf_HASENV)
+#define MP_HASENV_off(d)  (d->flags  &= ~MPf_HASENV)
+
+#define MP_DSTDERR(d)    (d->flags & MPf_DSTDERR)
+#define MP_DSTDERR_on(d)  (d->flags |= MPf_DSTDERR)
+#define MP_DSTDERR_off(d)  (d->flags  &= ~MPf_DSTDERR)
+
+#define MP_CLEANUP(d)    (d->flags & MPf_CLEANUP)
+#define MP_CLEANUP_on(d)  (d->flags |= MPf_CLEANUP)
+#define MP_CLEANUP_off(d)  (d->flags  &= ~MPf_CLEANUP)
+
+#define MP_RCLEANUP(d)    (d->flags & MPf_RCLEANUP)
+#define MP_RCLEANUP_on(d)  (d->flags |= MPf_RCLEANUP)
+#define MP_RCLEANUP_off(d)  (d->flags  &= ~MPf_RCLEANUP)
+
 #define PERL_GATEWAY_INTERFACE "CGI-Perl/1.1"
 /* Apache::SSI */
 #define PERL_APACHE_SSI_TYPE "text/x-perl-server-parsed-html"
@@ -78,6 +143,27 @@ typedef server_rec  * Apache__Server;
 #define PERL_SECTIONS
 #endif
 
+#ifndef API_EXPORT
+#define API_EXPORT(type)    type
+#endif
+
+#ifdef MULTITHREAD
+#include "multithread.h"
+extern void *mod_perl_mutex;
+#else
+#define mod_perl_mutex NULL 
+extern void *mod_perl_dummy_mutex;
+#endif
+
+#ifndef MULTITHREAD_H
+typedef void mutex;
+#define MULTI_OK (0)
+#define create_mutex(name)	((mutex *)mod_perl_dummy_mutex)
+#define acquire_mutex(mutex_id)	((int)MULTI_OK)
+#define release_mutex(mutex_id)	((int)MULTI_OK)
+
+#endif
+
 #ifdef PERL_STACKED_HANDLERS
 #define PERL_TAKE ITERATE
 #define PERL_CMD_INIT  Nullav
@@ -85,16 +171,31 @@ typedef server_rec  * Apache__Server;
 
 #define mod_perl_can_stack_handlers(sv) (SvTRUE(sv) && 1)
 
+/* always enable child_init for perl_init_ids */
+#if (MODULE_MAGIC_NUMBER >= 19970719) && !defined(WIN32)
+#define perl_init_ids
+# ifdef NO_PERL_CHILD_INIT
+#  undef NO_PERL_CHILD_INIT
+# endif
+# ifdef NO_PERL_CHILD_EXIT
+#  undef NO_PERL_CHILD_EXIT
+# endif
+#endif
+
+#ifndef perl_init_ids
+#define perl_init_ids mod_perl_init_ids()
+#endif
+
 #define PERL_CALLBACK(h,name) \
-perl_set_pid; \
+(void)acquire_mutex(mod_perl_mutex); \
 status = perl_run_stacked_handlers(h, r, Nullav); \
 if((status != OK) && (status != DECLINED)) { \
     MP_TRACE(fprintf(stderr, "%s handlers returned %d\n", h, status)); \
 } \
-else if(name != Nullav) { \
-    MP_TRACE(fprintf(stderr, "running server configured stacked handlers...\n")); \
+else if(AvTRUE(name)) { \
     status = perl_run_stacked_handlers(h, r, name); \
 } \
+(void)release_mutex(mod_perl_mutex); \
 MP_TRACE(fprintf(stderr, "%s handlers returned %d\n", h, status))
 
 
@@ -108,9 +209,12 @@ MP_TRACE(fprintf(stderr, "%s handlers returned %d\n", h, status))
 
 #define PERL_CALLBACK(h,name) \
 if(name != NULL) { \
-    SV *sv = newSVpv(name,0); \
+    SV *sv; \
+    (void)acquire_mutex(mod_perl_mutex); \
+    sv = newSVpv(name,0); \
     status = perl_call_handler(sv, r, Nullav); \
     SvREFCNT_dec(sv); \
+    (void)release_mutex(mod_perl_mutex); \
     MP_TRACE(fprintf(stderr, "perl_call %s '%s' returned: %d\n", h,name,status)); \
 } \
 else { \
@@ -119,8 +223,6 @@ else { \
 
 #endif
 
-#define PERL_EXIT_CLEANUP log_transaction(r);
-
 #if MODULE_MAGIC_NUMBER >= 19961007
 #define CHAR_P const char *
 #else
@@ -134,10 +236,17 @@ else { \
 #define PERL_READ_SETUP
 #endif 
 
+#if MODULE_MAGIC_NUMBER >= 19970622 
+#define PERL_SET_READ_LENGTH  r->read_length = 0
+#else
+#define PERL_SET_READ_LENGTH
+#endif 
+
 #if MODULE_MAGIC_NUMBER >= 19961125 
 #define PERL_READ_CLIENT \
 if(should_client_block(r)) { \
     nrd = get_client_block(r, buffer, bufsiz); \
+    PERL_SET_READ_LENGTH; \
 } 
 #else 
 #define PERL_READ_CLIENT \
@@ -185,8 +294,61 @@ PERL_READ_CLIENT
            table_set (env_arr, "GATEWAY_INTERFACE", PERL_GATEWAY_INTERFACE); \
        }
 
+/* on/off switches for callback hooks during server startup/shutdown */
+
+#ifndef NO_PERL_CHILD_INIT
+#define PERL_CHILD_INIT
+
+#define PERL_CHILD_INIT_HOOK perl_child_init
+
+#define PERL_CHILD_INIT_CMD_ENTRY \
+"PerlChildInitHandler", perl_cmd_child_init_handlers, \
+    NULL,	 \
+    RSRC_CONF, PERL_TAKE, "the Perl Child init handler routine name"  
+
+#define PERL_CHILD_INIT_CREATE(s) s->PerlChildInitHandler = PERL_CMD_INIT
+#else
+#define PERL_CHILD_INIT_HOOK NULL
+#define PERL_CHILD_INIT_CMD_ENTRY NULL
+#define PERL_CHILD_INIT_CREATE(s) 
+#endif
+
+#ifndef NO_PERL_CHILD_EXIT
+#define PERL_CHILD_EXIT
+
+#define PERL_CHILD_EXIT_HOOK perl_child_exit
+
+#define PERL_CHILD_EXIT_CMD_ENTRY \
+"PerlChildExitHandler", perl_cmd_child_exit_handlers, \
+    NULL,	 \
+    RSRC_CONF, PERL_TAKE, "the Perl Child exit handler routine name"  
+
+#define PERL_CHILD_EXIT_CREATE(s) s->PerlChildExitHandler = PERL_CMD_INIT
+#else
+#define PERL_CHILD_EXIT_HOOK NULL
+#define PERL_CHILD_EXIT_CMD_ENTRY NULL
+#define PERL_CHILD_EXIT_CREATE(s) 
+#endif
+
 /* on/off switches for callback hooks during request stages */
 
+#ifndef NO_PERL_POST_READ_REQUEST
+#define PERL_POST_READ_REQUEST
+
+#define PERL_POST_READ_REQUEST_HOOK perl_post_read_request
+
+#define PERL_POST_READ_REQUEST_CMD_ENTRY \
+"PerlPostReadRequestHandler", perl_cmd_post_read_request_handlers, \
+    NULL, \
+    RSRC_CONF, PERL_TAKE, "the Perl Post Read Request handler routine name" 
+
+#define PERL_POST_READ_REQUEST_CREATE(s) s->PerlPostReadRequestHandler = PERL_CMD_INIT
+#else
+#define PERL_POST_READ_REQUEST_HOOK NULL
+#define PERL_POST_READ_REQUEST_CMD_ENTRY NULL
+#define PERL_POST_READ_REQUEST_CREATE(s)
+#endif
+
 #ifndef NO_PERL_TRANS
 #define PERL_TRANS
 
@@ -204,6 +366,7 @@ PERL_READ_CLIENT
 #define PERL_TRANS_CREATE(s) 
 #endif
 
+
 #ifndef NO_PERL_AUTHEN
 #define PERL_AUTHEN
 
@@ -363,8 +526,11 @@ typedef struct {
     char **PerlModules;
     int  NumPerlModules;
     int  PerlTaintCheck;
-    PERL_CMD_TYPE   *PerlTransHandler;
     int  PerlWarn;
+    PERL_CMD_TYPE *PerlPostReadRequestHandler;
+    PERL_CMD_TYPE *PerlTransHandler;
+    PERL_CMD_TYPE *PerlChildInitHandler;
+    PERL_CMD_TYPE *PerlChildExitHandler;
 } perl_server_config;
 
 typedef struct {
@@ -379,53 +545,119 @@ typedef struct {
     PERL_CMD_TYPE *PerlHeaderParserHandler;
     PERL_CMD_TYPE *PerlInitHandler;
     table *env;
-    int has_env;
     table *vars;
-    int  sendheader;
-    int  new_sendheader;
-    int setup_env;
+    U32 flags;
 } perl_dir_config;
 
-extern module perl_module;
+#ifndef MODULE_VAR_EXPORT
+#define MODULE_VAR_EXPORT
+#endif
+
+extern module MODULE_VAR_EXPORT perl_module;
 
 /* a couple for -Wall sanity sake */
-int multi_log_transaction(request_rec *r);
 int basic_http_header(request_rec *r);
-/* prototypes */
+int translate_name (request_rec *);
+int log_transaction (request_rec *r);
+
+/* mod_perl prototypes */
+
+/* perlxsi.c */
+void xs_init (void);
+
+/* mod_perl.c */
+
+/* generic handler stuff */ 
 int perl_handler_ismethod(HV *class, char *sub);
-int perl_call_handler(SV *sv, request_rec *r, AV *args);
-int mod_perl_push_handlers(SV *self, SV *hook, SV *sub, AV *handlers);
+API_EXPORT(int) perl_call_handler(SV *sv, request_rec *r, AV *args);
+
+/* stacked handler stuff */
+int mod_perl_push_handlers(SV *self, char *hook, SV *sub, AV *handlers);
+SV *mod_perl_pop_handlers(SV *self, SV *hook);
+void *mod_perl_clear_handlers(SV *self, SV *hook);
+SV *mod_perl_fetch_handlers(SV *self, SV *hook);
 int perl_run_stacked_handlers(char *hook, request_rec *r, AV *handlers);
-int perl_handler(request_rec *r);
+
+/* plugin slots */
 void perl_startup(server_rec *s, pool *p);
-void *perl_merge_dir_config(pool *p, void *basev, void *addv);
-void *create_perl_dir_config(pool *p, char *dirname);
-void *create_perl_server_config(pool *p, server_rec *s);
+int perl_handler(request_rec *r);
+void perl_child_init(server_rec *, pool *);
+void perl_child_exit(server_rec *, pool *);
 int perl_translate(request_rec *r);
 int perl_authenticate(request_rec *r);
 int perl_authorize(request_rec *r);
 int perl_access(request_rec *r);
 int perl_type_checker(request_rec *r);
 int perl_fixup(request_rec *r);
+int perl_post_read_request(request_rec *r);
 int perl_logger(request_rec *r);
 int perl_header_parser(request_rec *r);
-CHAR_P perl_end_section (cmd_parms *cmd, void *dummy);
+int perl_hook(char *name);
+int PERL_RUNNING(void);
+
+/* per-request gunk */
+int mod_perl_sent_header(request_rec *r, int val);
+int mod_perl_seqno(SV *self);
+request_rec *perl_request_rec(request_rec *);
+void perl_setup_env(request_rec *r);
+SV  *perl_bless_request_rec(request_rec *); 
+void perl_set_request_rec(request_rec *); 
+void mod_perl_cleanup_handler(void *data);
+void mod_perl_end_cleanup(void *data);
+void mod_perl_register_cleanup(request_rec *r, SV *sv);
+
+/* perl_util.c */
+
+void perl_util_cleanup(void);
+void mod_perl_clear_rgy_endav(request_rec *r, SV *sv);
+void perl_run_rgy_endav(char *s);
+void perl_run_endav(char *s);
+void perl_call_halt(void);
+CV *empty_anon_sub(void);
+int perl_require_module(char *, server_rec *);
+void newCONSTSUB(HV *stash, char *name, SV *sv);
+void perl_clear_env(void);
+void mod_perl_init_ids(void);
+int perl_eval_ok(server_rec *);
+void perl_incpush(char *s);
+
+/* perlio.c */
+
+void perl_stdin2client(request_rec *);
+API_EXPORT(void) perl_stdout2client(request_rec *); 
+
+/* perl_config.c */
+
+void *perl_merge_dir_config(pool *p, void *basev, void *addv);
+void *perl_create_dir_config(pool *p, char *dirname);
+void *perl_create_server_config(pool *p, server_rec *s);
+
 CHAR_P perl_section (cmd_parms *cmd, void *dummy, CHAR_P arg);
+CHAR_P perl_end_section (cmd_parms *cmd, void *dummy);
+CHAR_P perl_limit_section(cmd_parms *cmd, void *dummy, HV *hv);
 CHAR_P perl_urlsection (cmd_parms *cmd, void *dummy, HV *hv);
+CHAR_P perl_dirsection (cmd_parms *cmd, void *dummy, HV *hv);
+CHAR_P perl_filesection (cmd_parms *cmd, void *dummy, HV *hv);
+void perl_add_file_conf (server_rec *s, void *url_config);
+void perl_handle_command(cmd_parms *cmd, void *dummy, char *line);
+void perl_handle_command_av(AV *av, I32 n, char *key, cmd_parms *cmd, void *dummy);
+
 CHAR_P perl_cmd_script (cmd_parms *parms, void *dummy, char *arg);
 CHAR_P perl_cmd_module (cmd_parms *parms, void *dummy, char *arg);
-CHAR_P perl_cmd_var(cmd_parms *cmd, void *rec, char *key, char *val);
-CHAR_P perl_cmd_setenv(cmd_parms *cmd, void *rec, char *key, char *val);
-CHAR_P perl_cmd_sendheader (cmd_parms *cmd, void *rec, int arg);
-CHAR_P perl_cmd_new_sendheader (cmd_parms *cmd, void *rec, int arg);
+CHAR_P perl_cmd_var(cmd_parms *cmd, perl_dir_config *rec, char *key, char *val);
+CHAR_P perl_cmd_setenv(cmd_parms *cmd, perl_dir_config *rec, char *key, char *val);
+CHAR_P perl_cmd_env (cmd_parms *cmd, perl_dir_config *rec, int arg);
+CHAR_P perl_cmd_sendheader (cmd_parms *cmd, perl_dir_config *rec, int arg);
 CHAR_P perl_cmd_tainting (cmd_parms *parms, void *dummy, int arg);
 CHAR_P perl_cmd_warn (cmd_parms *parms, void *dummy, int arg);
-CHAR_P perl_cmd_env (cmd_parms *cmd, void *rec, int arg);
-void mod_perl_dir_env(perl_dir_config *cld);
+
 CHAR_P perl_cmd_init_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
 CHAR_P perl_cmd_cleanup_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
 CHAR_P perl_cmd_header_parser_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
+CHAR_P perl_cmd_post_read_request_handlers (cmd_parms *parms, void *dumm, char *arg);
 CHAR_P perl_cmd_trans_handlers (cmd_parms *parms, void *dumm, char *arg);
+CHAR_P perl_cmd_child_init_handlers (cmd_parms *parms, void *dumm, char *arg);
+CHAR_P perl_cmd_child_exit_handlers (cmd_parms *parms, void *dumm, char *arg);
 CHAR_P perl_cmd_authen_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
 CHAR_P perl_cmd_authz_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
 CHAR_P perl_cmd_access_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
@@ -434,23 +666,4 @@ CHAR_P perl_cmd_fixup_handlers (cmd_parms *parms, perl_dir_config *rec, char *ar
 CHAR_P perl_cmd_handler_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
 CHAR_P perl_cmd_log_handlers (cmd_parms *parms, perl_dir_config *rec, char *arg);
 
-#ifdef APACHE_SSL
-void xs_init (void);
-#else
-void xs_init _((void));
-#endif
-int mod_perl_sent_header(SV *self, int val);
-int mod_perl_seqno(void);
-int perl_hook(char *name);
-request_rec *perl_request_rec(request_rec *);
-void perl_stdin2client(request_rec *);
-void perl_stdout2client(request_rec *); 
-int perl_require_module(char *, server_rec *);
-int  perl_eval_ok(server_rec *);
-void perl_setup_env(request_rec *r);
-SV  *perl_bless_request_rec(request_rec *); 
-void perl_set_request_rec(request_rec *); 
-void mod_perl_cleanup_handler(void *data);
-void mod_perl_end_cleanup(void *data);
-void mod_perl_register_cleanup(request_rec *r, SV *sv);
-int PERL_RUNNING(void);
+void mod_perl_dir_env(perl_dir_config *cld);
