@@ -1,19 +1,24 @@
 use Apache ();
 use strict;
-use vars qw($r);
-sub test { $r->print(sprintf "%s", $_[1] ? "ok $_[0]\n" : "not ok $_[0]\n") }
 
-my $tests = 21;
+my $tests = 24;
 my $i;
-$r = Apache->request;
-$r->content_type("text/html");
+my $r = Apache->request;
+$r->content_type("text/plain");
 $r->send_http_header;
 $r->print("1..$tests\n");
 
+sub test { 
+    Apache->request->
+	print(sprintf "%s", $_[1] ? "ok $_[0]\n" : "not ok $_[0]\n");
+}
+
 %ENV = $r->cgi_env;
 
+test ++$i, $r->filename eq $0;
+
 test ++$i, $ENV{GATEWAY_INTERFACE};
-test ++$i, $r->seqno;
+test ++$i, defined($r->seqno);
 test ++$i, $r->protocol;
 #hostname
 test ++$i, $r->status;
@@ -54,6 +59,9 @@ test ++$i, $r->handler;
 my $c = $r->connection;
 test ++$i, $c;
 test ++$i, $c->remote_ip;
+test ++$i, $c->remote_addr;
+test ++$i, $c->local_addr;
+
 #Connection::remote_host
 #Connection::remote_logname
 #Connection::user
@@ -64,3 +72,10 @@ test ++$i, $s;
 test ++$i, $s->server_admin;
 test ++$i, $s->server_hostname;
 test ++$i, $s->port;
+
+
+
+
+
+
+
